import os

import jax
from jax import random, numpy as jnp

# Force 2 jax  hosts
os.environ["XLA_FLAGS"] = "--xla_force_host_platform_device_count=2"

from jaxns import TerminationCondition
from jaxns.utils import sample_evidence
import numpy as np


def test_nested_sampling_run_results(all_run_results):
    for name, (log_Z_true, results) in all_run_results:
        print(f"Checking {name}")
        # Use numpy testing

        assert not np.isnan(results.log_Z_mean)
        assert not np.isnan(results.log_Z_uncert)

        log_Z_samples = sample_evidence(random.PRNGKey(42),
                                        results.num_live_points_per_sample,
                                        results.log_L_samples,
                                        S=1000)

        # Filter outliers
        select_mask = jnp.bitwise_and(
            log_Z_samples > jnp.percentile(log_Z_samples, 5),
            log_Z_samples < jnp.percentile(log_Z_samples, 95)
        )
        log_Z_samples = log_Z_samples[select_mask]

        import pylab as plt
        plt.hist(log_Z_samples, bins='auto')
        plt.show()
        log_Z_ensemble_mean = jnp.mean(log_Z_samples)
        log_Z_ensemble_std = jnp.std(log_Z_samples)
        np.testing.assert_allclose(log_Z_ensemble_mean, log_Z_true, atol=3.0 * results.log_Z_uncert)
        np.testing.assert_allclose(results.log_Z_mean, log_Z_ensemble_mean, atol=3. * results.log_Z_uncert)
        np.testing.assert_allclose(results.log_Z_uncert, log_Z_ensemble_std,
                                   atol=np.sqrt(results.log_Z_uncert ** 2 + log_Z_ensemble_std ** 2))


def test_gh108():
    import tensorflow_probability.substrates.jax as tfp
    from jax import random

    from jaxns import DefaultNestedSampler
    from jaxns import Model
    from jaxns import Prior

    import psutil
    import os
    tfpd = tfp.distributions

    def nested_sampling(key):

        def log_likelihood(theta):
            return 0.

        def prior_model():
            x = yield Prior(tfpd.Uniform(0., 1.))
            return x

        model = Model(prior_model=prior_model, log_likelihood=log_likelihood)

        ns = DefaultNestedSampler(model=model, max_samples=1e4)

        termination_reason, state = ns(key, term_cond=TerminationCondition())
        results = ns.to_results(termination_reason=termination_reason, state=state, trim=False)

    pid = os.getpid()
    python_process = psutil.Process(pid)

    ram_py = []
    jax.clear_caches()
    for i in range(3):
        nested_sampling(random.PRNGKey(i))
        jax.clear_caches()
        ram_py.append(python_process.memory_info()[0] / 2 ** 30)
        # print(ram_py[-1])

    # plt.plot(ram_py, 'k.-')
    # plt.xlabel('runs', fontsize=12)
    # plt.ylabel('python RAM usage(GB)', fontsize=12)
    # plt.show()

    np.testing.assert_allclose(ram_py, ram_py[0], atol=2e-3)

    ns_compile = jax.jit(nested_sampling).lower(random.PRNGKey(0)).compile()

    ram_py = []
    ram_py.append(python_process.memory_info()[0] / 2 ** 30)
    for i in range(3):
        ns_compile(random.PRNGKey(i))
        ram_py.append(python_process.memory_info()[0] / 2 ** 30)
        # print(ram_py[-1])

    np.testing.assert_allclose(ram_py, ram_py[0], atol=1e-6)



#!/usr/bin/env python

from setuptools import find_packages
from setuptools import setup

install_requires = [
    'jax',
    'jaxlib',
    'chex',
    'typing_extensions',
    'matplotlib',
    'numpy',
    'scipy',
    'tensorflow_probability',
    'tqdm',
    'dm-haiku',
    'optax',
    'jaxopt'
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='jaxns',
      version='2.4.8',
      description='Nested Sampling in JAX',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/joshuaalbert/jaxns",
      author='Joshua G. Albert',
      author_email='albert@strw.leidenuniv.nl',
      install_requires=install_requires,
      tests_require=[
          'pytest>=2.8',
      ],
      package_dir={'': './'},
      packages=find_packages('./'),
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
      ],
      python_requires='>=3.9',
      )



from typing import Union

import jax.numpy as jnp
from jax import random, vmap, lax
from jax._src.scipy.special import logsumexp
from jax.scipy.stats import multivariate_normal


def initialize_params(key, data, n_components: int):
    n, d = data.shape

    # Initialize means by selecting random data points
    assign_idx = random.choice(key, n, shape=(n_components,), replace=False)
    means = data[assign_idx]

    # Initialize covariances as the empirical covariance of the data
    # cov = jnp.cov(data, rowvar=False)
    cov = jnp.diag(jnp.var(data, axis=0))
    covariances = jnp.repeat(cov[None, ...], n_components, axis=0)

    # Initialize mixture weights uniformly
    log_weights = jnp.full((n_components,), -jnp.log(n_components))

    return means, covariances, log_weights


def e_step(data, means, covariances, log_weights, mask):
    n, d = data.shape
    n_components = means.shape[0]

    # Compute the probabilities of each data point belonging to each Gaussian
    logpdf = vmap(lambda m, c: multivariate_normal.logpdf(data, m, c))(means, covariances)  # num_clusters, num_data
    if mask is not None:
        logpdf = jnp.where(mask[None, :], logpdf, -jnp.inf)
    logpdf_weighted = logpdf + log_weights[:, None]
    # Normalize probabilities
    log_responsibilities = logpdf_weighted - logsumexp(logpdf_weighted, axis=0)
    return log_responsibilities


def m_step(data, log_responsibilities):
    n_components, num_data = log_responsibilities.shape
    _, d = data.shape

    # Update means, covariances, and weights
    log_weights = logsumexp(log_responsibilities, axis=1) - jnp.log(num_data)  # num_components

    # num_components, num_data X num_data, D -> num_components, D
    weighted_responsibilities = jnp.exp(log_responsibilities - log_weights[:, None] - jnp.log(num_data))
    means = jnp.matmul(weighted_responsibilities, data)

    centered_data = data[None, :, :] - means[:, None, :]  # num_components, num_data, D

    covariances = jnp.einsum("cn,cnd,cne->cde", weighted_responsibilities, centered_data, centered_data)
    covariances = covariances + 1e-4 * jnp.eye(d)
    return means, covariances, log_weights


# No invariance under jit...
def em_gmm(key, data, n_components, mask: Union[jnp.ndarray, None] = None, n_iters=10, tol=1e-6):
    means, covariances, log_weights = initialize_params(key, data, n_components)
    params = (means, covariances, log_weights)

    def body(state):
        _, i, params = state
        log_responsibilities = e_step(data, *params, mask=mask)
        new_params = m_step(data, log_responsibilities)
        done = False
        for param, new_param in zip(params, new_params):
            done = done | (jnp.all(jnp.abs(jnp.array(param) - jnp.array(new_param)) < tol)) | (i >= n_iters)

        return done, i + 1, new_params

    def cond(state):
        done, _, params = state
        return jnp.bitwise_not(done)

    _, total_iters, params = lax.while_loop(
        cond,
        body,
        (jnp.asarray(False), jnp.asarray(0), params)
    )

    cluster_id = jnp.argmax(e_step(data, *params, mask=mask), axis=0)
    return cluster_id, params, total_iters



import os
from time import monotonic_ns

import jax

# Force 2 jax  hosts
os.environ["XLA_FLAGS"] = "--xla_force_host_platform_device_count=2"

import pytest
from jax import numpy as jnp, random
from jax._src.scipy.linalg import solve_triangular
from tensorflow_probability.substrates import jax as tfp

from jaxns.framework.bases import PriorModelGen
from jaxns.framework.model import Model
from jaxns.framework.prior import Prior
from jaxns.nested_sampler.standard_static import StandardStaticNestedSampler
from jaxns.public import DefaultNestedSampler
from jaxns.samplers.multi_ellipsoidal_samplers import MultiEllipsoidalSampler
from jaxns.internals.types import TerminationCondition
from jaxns.utils import bruteforce_evidence, summary

# from jaxns.nested_sampler import ApproximateNestedSampler, ExactNestedSampler

tfpd = tfp.distributions


class Timer:
    def __enter__(self):
        self.t0 = monotonic_ns()

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(f"Time to execute: {(monotonic_ns() - self.t0) / 1e9} seconds.")


@pytest.fixture(scope='package')
def basic_model():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1), name='x')
        return x

    def log_likelihood(x):
        return - jnp.sum(x ** 2)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)

    log_Z_true = bruteforce_evidence(model=model, S=200)
    return model, log_Z_true


@pytest.fixture(scope='package')
def basic_run_results(basic_model):
    model, log_Z_true = basic_model

    ns = DefaultNestedSampler(model=model, max_samples=1000, verbose=True)
    ns_jit = jax.jit(lambda key: ns(key))
    ns_compiled = ns_jit.lower(random.PRNGKey(42)).compile()
    with Timer():
        termination_reason, state = ns_compiled(random.PRNGKey(42))
        termination_reason.block_until_ready()
    results = ns.to_results(termination_reason=termination_reason, state=state)
    ns.plot_diagnostics(results)
    ns.summary(results)
    # exact_ns.plot_cornerplot(results)
    return log_Z_true, results


@pytest.fixture(scope='package')
def basic2_model():
    n = 2

    # Prior is uniform in U[0,1]
    # Likelihood is 1 - x**n
    # Z = 1 - 1/n+1

    log_Z_true = jnp.log(1. - 1. / (n + 1))

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1))
        return x

    def log_likelihood(x):
        return jnp.log(1. - x ** n)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    return model, log_Z_true


@pytest.fixture(scope='package')
def basic2_run_results(basic2_model):
    model, log_Z_true = basic2_model

    ns = DefaultNestedSampler(model=model, max_samples=1000, verbose=True)

    ns_jit = jax.jit(lambda key: ns(key))
    ns_compiled = ns_jit.lower(random.PRNGKey(42)).compile()
    with Timer():
        termination_reason, state = ns_compiled(random.PRNGKey(42))
        termination_reason.block_until_ready()
    results = ns.to_results(termination_reason=termination_reason, state=state)
    # exact_ns.plot_diagnostics(results)
    ns.summary(results)
    # exact_ns.plot_cornerplot(results)
    return log_Z_true, results


@pytest.fixture(scope='package')
def basic3_model():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2), name='x')
        y = yield Prior(tfpd.Normal(loc=2, scale=x), name='y')
        z = x + y
        return z

    def log_likelihood(z):
        return -z ** 2

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)

    log_Z_true = bruteforce_evidence(model=model, S=500)
    return model, log_Z_true


@pytest.fixture(scope='package')
def basic3_run_results(basic3_model):
    model, log_Z_true = basic3_model

    ns = DefaultNestedSampler(model=model, max_samples=2000, k=0, verbose=True)

    ns_jit = jax.jit(lambda key: ns(key))
    ns_compiled = ns_jit.lower(random.PRNGKey(42)).compile()
    with Timer():
        termination_reason, state = ns_compiled(random.PRNGKey(42))
        termination_reason.block_until_ready()
    results = ns.to_results(termination_reason=termination_reason, state=state)
    ns.plot_diagnostics(results)
    ns.summary(results)
    ns.plot_cornerplot(results)
    return log_Z_true, results


@pytest.fixture(scope='package')
def plateau_model():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1))
        return x

    def log_likelihood(x):
        return 0.

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)

    log_Z_true = jnp.asarray(0.)
    return model, log_Z_true


@pytest.fixture(scope='package')
def plateau_run_results(plateau_model):
    model, log_Z_true = plateau_model
    ns = DefaultNestedSampler(
        model=model,
        max_samples=1000, verbose=True
    )

    ns_jit = jax.jit(lambda key: ns(key))
    ns_compiled = ns_jit.lower(random.PRNGKey(42)).compile()
    with Timer():
        termination_reason, state = ns_compiled(random.PRNGKey(42))
        termination_reason.block_until_ready()
    results = ns.to_results(termination_reason=termination_reason, state=state)
    # exact_ns.plot_diagnostics(results)
    ns.summary(results)
    # exact_ns.plot_cornerplot(results)

    return log_Z_true, results


@pytest.fixture(scope='package')
def basic_mvn_model():
    def log_normal(x, mean, cov):
        L = jnp.linalg.cholesky(cov)
        dx = x - mean
        dx = solve_triangular(L, dx, lower=True)
        return -0.5 * x.size * jnp.log(2. * jnp.pi) - jnp.sum(jnp.log(jnp.diag(L))) \
            - 0.5 * dx @ dx

    ndims = 8
    prior_mu = 15 * jnp.ones(ndims)
    prior_cov = jnp.diag(jnp.ones(ndims)) ** 2

    data_mu = jnp.zeros(ndims)
    data_cov = jnp.diag(jnp.ones(ndims)) ** 2
    data_cov = jnp.where(data_cov == 0., 0.99, data_cov)

    log_Z_true = log_normal(data_mu, prior_mu, prior_cov + data_cov)
    # not super happy with this being 1.58 and being off by like 0.1. Probably related to the ESS.
    post_mu = prior_cov @ jnp.linalg.inv(prior_cov + data_cov) @ data_mu + data_cov @ jnp.linalg.inv(
        prior_cov + data_cov) @ prior_mu

    print(f"True post mu:{post_mu}")
    print(f"True log Z: {log_Z_true}")

    def prior_model() -> PriorModelGen:
        x = yield Prior(
            tfpd.MultivariateNormalTriL(loc=prior_mu, scale_tril=jnp.linalg.cholesky(prior_cov)),
            name='x')
        return x

    def log_likelihood(x):
        return tfpd.MultivariateNormalTriL(loc=data_mu, scale_tril=jnp.linalg.cholesky(data_cov)).log_prob(x)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    return log_Z_true, model


@pytest.fixture(scope='package')
def basic_mvn_run_results(basic_mvn_model):
    log_Z_true, model = basic_mvn_model

    ns = DefaultNestedSampler(model=model, max_samples=100000, verbose=True)

    ns_jit = jax.jit(lambda key: ns(key))
    ns_compiled = ns_jit.lower(random.PRNGKey(42)).compile()
    with Timer():
        termination_reason, state = ns_compiled(random.PRNGKey(42))
        termination_reason.block_until_ready()
    results = ns.to_results(termination_reason=termination_reason, state=state)
    ns.plot_diagnostics(results)
    ns.summary(results)
    # exact_ns.plot_cornerplot(results)

    return log_Z_true, results


@pytest.fixture(scope='package')
def basic_mvn_run_results_parallel(basic_mvn_model):
    log_Z_true, model = basic_mvn_model

    ns = DefaultNestedSampler(model=model, max_samples=1e5, num_parallel_workers=2, verbose=True)

    ns_jit = jax.jit(lambda key: ns(key))
    ns_compiled = ns_jit.lower(random.PRNGKey(42)).compile()
    with Timer():
        termination_reason, state = ns_compiled(random.PRNGKey(42))
        termination_reason.block_until_ready()
    results = ns.to_results(termination_reason=termination_reason, state=state)
    ns.plot_diagnostics(results)
    ns.summary(results)
    # exact_ns.plot_cornerplot(results)

    return log_Z_true, results


@pytest.fixture(scope='package')
def multiellipsoidal_mvn_run_results(basic_mvn_model):
    log_Z_true, model = basic_mvn_model

    # model.sanity_check(random.PRNGKey(42), S=100)
    ns = StandardStaticNestedSampler(
        init_efficiency_threshold=0.1,
        model=model,
        num_live_points=model.U_ndims * 20,
        num_parallel_workers=1,
        max_samples=50000,
        sampler=MultiEllipsoidalSampler(model=model, depth=0, expansion_factor=2.0),
        verbose=True
    )
    ns_jit = jax.jit(lambda key: ns._run(key, term_cond=TerminationCondition()))
    ns_compiled = ns_jit.lower(random.PRNGKey(42)).compile()
    with Timer():
        termination_reason, state = ns_compiled(random.PRNGKey(42))
        termination_reason.block_until_ready()
    results = ns._to_results(termination_reason=termination_reason, state=state, trim=True)
    # plot_diagnostics(results)
    summary(results)
    # plot_cornerplot(results)

    return log_Z_true, results


@pytest.fixture(scope='package')
def all_run_results(
        basic_run_results,
        basic2_run_results,
        basic3_run_results,
        plateau_run_results,
        basic_mvn_run_results,
        basic_mvn_run_results_parallel,
        multiellipsoidal_mvn_run_results
):
    # Return tuples with names
    return [
        ('basic', basic_run_results),
        ('basic2', basic2_run_results),
        ('basic3', basic3_run_results),
        ('plateau', plateau_run_results),
        ('basic_mvn', basic_mvn_run_results),
        ('basic_mvn_parallel', basic_mvn_run_results_parallel),
        ('multiellipsoidal_mvn', multiellipsoidal_mvn_run_results)
    ]



import tensorflow_probability.substrates.jax as tfp

from jaxns import Prior, Model
from jaxns.experimental import EvidenceMaximisation

tfpd = tfp.distributions


def test_basic():
    def prior_model():
        x = yield Prior(tfpd.Uniform(0., 1.))
        y = yield Prior(tfpd.Normal(x, 1.), name='y').parametrised()
        sigma = yield Prior(tfpd.Exponential(1.))
        return y, sigma

    def log_likelihood(y, sigma):
        return tfpd.Normal(y, sigma).log_prob(0.)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)

    em = EvidenceMaximisation(model=model, ns_kwargs=dict(max_samples=1e5))
    ns_results, params = em.train(num_steps=10)



from typing import Tuple, Union, Optional, Literal

import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp, vmap, lax
from jax._src.scipy.special import gammaln
from tensorflow_probability.substrates.jax.math import lbeta, betaincinv

from jaxns.internals.log_semiring import cumulative_logsumexp
from jaxns.framework.bases import BaseAbstractPrior
from jaxns.internals.types import FloatArray, IntArray, BoolArray, float_type

tfpd = tfp.distributions

__all__ = [
    "Bernoulli",
    "Beta",
    "Categorical",
    "ForcedIdentifiability",
    "Poisson",
    "UnnormalisedDirichlet"
]


class Bernoulli(BaseAbstractPrior):
    def __init__(self, *, logits=None, probs=None, name: Optional[str] = None):
        super(Bernoulli, self).__init__(name=name)
        self.dist = tfpd.Bernoulli(logits=logits, probs=probs)

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self) -> Tuple[int, ...]:
        return self._shape()

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self.dist.cdf(X)

    def _log_prob(self, X) -> FloatArray:
        return self.dist.log_prob(X)

    def _quantile(self, U):
        probs = self.dist._probs_parameter_no_checks()
        sample = jnp.less(U, probs)
        return sample.astype(self.dtype)


class Beta(BaseAbstractPrior):
    def __init__(self, *, concentration0=None, concentration1=None, name: Optional[str] = None):
        super(Beta, self).__init__(name=name)
        self.dist = tfpd.Beta(concentration0=concentration0, concentration1=concentration1)

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self) -> Tuple[int, ...]:
        return self._shape()

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self.dist.cdf(X)

    def _log_prob(self, X) -> FloatArray:
        return self.dist.log_prob(X)

    def _quantile(self, U):
        alpha = self.dist.concentration0
        beta = self.dist.concentration1
        # cdf(x, a, b) = I(x,a,b) = B(x,a,b)/B(a,b)
        # G = B(a,b) cdf(x, a, b) = B(a,b) I(x,a,b) = B(x,a,b)
        # quantile(u, a, b) = B^{-1}(B(a,b) U, a, b)
        Y = jnp.exp(lbeta(alpha, beta) + jnp.log(U))
        X = betaincinv(alpha, beta, Y)
        return X.astype(self.dtype)


class Categorical(BaseAbstractPrior):
    def __init__(self, parametrisation: Literal['gumbel_max', 'cdf'], *, logits=None, probs=None,
                 name: Optional[str] = None):
        """
        Initialised Categorical special prior.

        Args:
            parametrisation: 'cdf' is good for discrete params with correlation between neighbouring categories,
                otherwise gumbel is better.
            logits: log-prob of each category
            probs: prob of each category
            name: optional name
        """
        super(Categorical, self).__init__(name=name)
        self.dist = tfpd.Categorical(logits=logits, probs=probs)
        self._parametrisation = parametrisation

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self) -> Tuple[int, ...]:
        if self._parametrisation == 'gumbel_max':
            return self._shape() + (self.dist._num_categories(),)
        elif self._parametrisation == 'cdf':
            return self._shape()

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        if self._parametrisation == 'gumbel_max':
            return self._quantile_gumbelmax(U)
        elif self._parametrisation == 'cdf':
            return self._quantile_cdf(U)

    def _inverse(self, X) -> FloatArray:
        return self.dist.cdf(X)

    def _log_prob(self, X) -> FloatArray:
        return self.dist.log_prob(X)

    def _quantile_gumbelmax(self, U):
        logits = self.dist._logits_parameter_no_checks()
        sample_dtype = self.dtype
        z = -jnp.log(-jnp.log(U))  # gumbel
        draws = jnp.argmax(logits + z, axis=-1).astype(sample_dtype)
        return draws

    def _quantile_cdf(self, U):
        """
        The quantile for CDF parametrisation.

        Args:
            U: [...]

        Returns:
            [...]
        """
        logits = self.dist._logits_parameter_no_checks()  # [..., N]
        cum_logits = cumulative_logsumexp(logits, axis=-1)  # [..., N]
        cum_logits -= cum_logits[..., -1:]
        N = cum_logits.shape[-1]
        cum_logits_flat = jnp.reshape(cum_logits, (-1, N))
        log_U_flat = jnp.reshape(jnp.log(U), (-1,))

        category_flat = vmap(lambda a, v: jnp.searchsorted(a, v, side='left'))(cum_logits_flat, log_U_flat)
        category = jnp.reshape(category_flat, U.shape)
        return category


class ForcedIdentifiability(BaseAbstractPrior):
    """
    Prior for a sequence of `n` random variables uniformly distributed on U[low, high] such that U[i,...] <= U[i+1,...].
    For broadcasting the resulting random variable is sorted on the first dimension elementwise.

    Args:
        n: number of samples within [low,high]
        low: minimum of distribution
        high: maximum of distribution
    """

    def __init__(self, *, n: int, low=None, high=None, name: Optional[str] = None):
        super(ForcedIdentifiability, self).__init__(name=name)
        self.n = n
        low, high = jnp.broadcast_arrays(low, high)
        self.low = low
        self.high = high

    def _dtype(self):
        return float_type

    def _base_shape(self) -> Tuple[int, ...]:
        return (self.n,) + jnp.shape(self.low)

    def _shape(self) -> Tuple[int, ...]:
        return (self.n,) + jnp.shape(self.low)

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self._cdf(X)

    def _log_prob(self, X) -> FloatArray:
        log_n_fac = gammaln(self.n + 1)
        diff = self.high - self.low
        log_prob = - log_n_fac - self.n * jnp.log(diff)
        # no check that X is inside high and low
        return log_prob

    def _cdf(self, X):
        log_theta = jnp.log((X - self.low) / (self.high - self.low))  # [n, ...]

        def inv_body(state, X):
            (log_theta_prev,) = state
            (log_theta, i) = X
            log_x = i * (log_theta - log_theta_prev)
            return (log_theta,), (log_x,)

        # Initial log_x value
        log_init_x = jnp.zeros(self.shape[1:], self.dtype)  # [...]
        _, (log_x,) = lax.scan(inv_body, (log_init_x,), (log_theta, jnp.arange(1, self.n + 1)), reverse=True)
        U = jnp.exp(log_x)
        return U.astype(self.dtype)

    def _quantile(self, U):
        log_x = jnp.log(U)  # [n, ...]

        # theta[i] = theta[i-1] * (1 - x[i]) + theta_max * x[i]
        def body(state, X):
            (log_theta,) = state
            (log_x, i) = X
            log_theta = log_x / i + log_theta  # shrinkage
            return (log_theta,), (log_theta,)

        log_init_theta = jnp.zeros(self.shape[1:], self.dtype)  # [...] -- log(1)
        _, (log_theta,) = lax.scan(body, (log_init_theta,), (log_x, jnp.arange(1, self.n + 1)), reverse=True)
        theta = self.low + (self.high - self.low) * jnp.exp(log_theta)
        return theta.astype(self.dtype)


class Poisson(BaseAbstractPrior):
    def __init__(self, *, rate=None, log_rate=None, name: Optional[str] = None):
        super(Poisson, self).__init__(name=name)
        self.dist = tfpd.Poisson(rate=rate, log_rate=log_rate)

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self) -> Tuple[int, ...]:
        return self._shape()

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self.dist.batch_shape_tensor()) + tuple(self.dist.event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self.dist.cdf(X)

    def _log_prob(self, X) -> FloatArray:
        return self.dist.log_prob(X)

    def _quantile(self, U):
        """
        Algorithmic Poisson generator based upon the inversion by sequential search

        .. code-block::

                init:
                    Let log_x ← -inf, log_p ← −λ, log_s ← log_p.
                    Generate uniform random number u in [0,1].
                while log_u > log_s do:
                    log_x ← logaddexp(log_x, 0).
                    log_p ← log_p + log_λ - log_x.
                    log_s ← logaddexp(log_s, log_p).
                return exp(log_x).
        """
        log_rate = self.dist.log_rate_parameter()
        rate = self.dist.rate_parameter()
        tiny = 1e-7  # For anything smaller it never terminates
        U = jnp.minimum(U, jnp.asarray(1., U.dtype) - tiny)
        log_u = jnp.log(U)
        log_x = jnp.full(log_u.shape, -jnp.inf)
        log_p = -rate
        log_s = log_p

        def body(state):
            (_log_x, _log_p, _log_s) = state
            impute = log_u <= _log_s
            log_x = jnp.logaddexp(_log_x, 0.)
            log_p = _log_p + log_rate - log_x
            log_s = jnp.logaddexp(_log_s, log_p)

            log_x = jnp.where(impute, _log_x, log_x)
            log_p = jnp.where(impute, _log_p, log_p)
            log_x = jnp.where(impute, _log_x, log_x)

            return (log_x, log_p, log_s)

        (log_x, log_p, log_s) = lax.while_loop(lambda s: jnp.any(log_u > s[2]),
                                               body,
                                               (log_x, log_p, log_s))
        return jnp.exp(log_x).astype(self.dtype)


class UnnormalisedDirichlet(BaseAbstractPrior):
    """
    Represents an unnormalised dirichlet distribution of K classes.
    That is, the output is related to the K-simplex via normalisation.

    X ~ UnnormalisedDirichlet(alpha)
    Y = X / sum(X) ==> Y ~ Dirichlet(alpha)
    """

    def __init__(self, *, concentration, name: Optional[str] = None):
        super(UnnormalisedDirichlet, self).__init__(name=name)
        self._dirichlet_dist = tfpd.Dirichlet(concentration=concentration)
        self._gamma_dist = tfpd.Gamma(concentration=concentration, log_rate=0.)

    def _dtype(self):
        return self._dirichlet_dist.dtype

    def _base_shape(self) -> Tuple[int, ...]:
        return self._shape()

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self._dirichlet_dist.batch_shape_tensor()) + tuple(self._dirichlet_dist.event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._quantile(U)

    def _inverse(self, X) -> FloatArray:
        return self._gamma_dist.cdf(X)

    def _log_prob(self, X) -> FloatArray:
        return jnp.sum(self._gamma_dist.log_prob(X), axis=-1)

    def _quantile(self, U):
        gamma = self._gamma_dist.quantile(U).astype(self.dtype)
        return gamma



from typing import TypeVar, Callable, Tuple, Optional

from jax import lax, numpy as jnp, tree_map, tree_util

from jaxns.internals.types import IntArray, int_type

V = TypeVar('V')
Y = TypeVar('Y')


def cumulative_op_static(op: Callable[[V, Y], V], init: V, xs: Y, pre_op: bool = False, unroll: int = 1) -> Tuple[
    V, V]:
    """
    Compute a cumulative operation on a list of values.

    Args:
        op: the operation to perform
        init: the initial value
        xs: the list of values
        pre_op: if True, the operation is applied before the accumulation, so the first value is the initial value.
        unroll: how many iterations to unroll the loop at a time

    Returns:
        the final accumulated value, and the result of the cumulative operation applied on input
    """

    def body(accumulate: V, y: Y):
        next_accumulate = op(accumulate, y)
        if pre_op:
            return next_accumulate, accumulate
        return next_accumulate, next_accumulate

    final_accumulate, result = lax.scan(
        f=body,
        init=init,
        xs=xs,
        unroll=unroll
    )

    return final_accumulate, result


def cumulative_op_dynamic(op: Callable[[V, Y], V], init: V, xs: Y, stop_idx: IntArray, pre_op: bool = False,
                          empty_fill: Optional[V] = None) -> Tuple[
    V, V]:
    """
    Compute a cumulative operation on a list of values with a dynamic stop index.

    Args:
        op: the operation to perform
        init: the initial value
        xs: the list of values
        stop_idx: how many accumulations to perform
        pre_op: if True, the operation is applied before the accumulation, so the first value is the initial value.
        empty_fill: the value to fill the output with if the stop_idx is provided, else uses `init`

    Returns:
        the final accumulated value, and the result of the cumulative operation applied on input
    """

    def cond(carry: Tuple[V, IntArray, V]):
        (accumulate, i, output) = carry
        return jnp.less(i, stop_idx)

    def body(carry: Tuple[V, IntArray, V]):
        (accumulate, i, output) = carry
        y = tree_map(lambda x: x[i], xs)
        next_accumulate = op(accumulate, y)
        next_i = i + jnp.ones_like(i)
        if pre_op:
            next_output = tree_map(lambda a, b: a.at[i].set(b), output, accumulate)
            return (next_accumulate, next_i, next_output)
        next_output = tree_map(lambda a, b: a.at[i].set(b), output, next_accumulate)
        return (next_accumulate, next_i, next_output)

    length = tree_util.tree_flatten(xs)[0][0].shape[0]

    output = tree_map(
        lambda x: jnp.tile(x[None], [length] + [1] * len(x.shape)),
        empty_fill if empty_fill is not None else init
    )

    w_init = (init, jnp.asarray(0, int_type), output)

    (final_accumulate, _, final_output) = lax.while_loop(
        cond_fun=cond,
        body_fun=body,
        init_val=w_init
    )

    return final_accumulate, final_output



from jax import numpy as jnp, random

from jaxns.internals.log_semiring import LogSpace, signed_logaddexp, cumulative_logsumexp, logaddexp, \
    is_complex
from jaxns.internals.types import complex_type


def test_log_space():
    #
    a = 5.
    b = 10.

    _a = LogSpace(jnp.log(a))
    _b = LogSpace(jnp.log(b))

    assert jnp.isclose((_a + _b).value, a + b)
    assert jnp.isclose((_a - _b).value, a - b)
    assert jnp.isclose((_a * _b).value, a * b)
    assert jnp.isclose((_a / _b).value, a / b)
    assert jnp.isclose((_a ** 2).value, a ** 2)
    assert jnp.isclose((_a ** 3).value, a ** 3)
    assert jnp.isclose((_a ** 0.5).value, a ** 0.5)

    #
    a = 0.
    b = 10.

    _a = LogSpace(jnp.log(a))
    _b = LogSpace(jnp.log(b))

    assert jnp.isclose((_a + _b).value, a + b)
    assert jnp.isclose((_a - _b).value, a - b)
    assert jnp.isclose((_a * _b).value, a * b)
    assert jnp.isclose((_a / _b).value, a / b)
    assert jnp.isclose((_a ** 2).value, a ** 2)
    assert jnp.isclose((_a ** 3).value, a ** 3)

    a = -5.
    b = 10.

    _a = LogSpace(jnp.log(jnp.abs(a)), jnp.sign(a))
    _b = LogSpace(jnp.log(jnp.abs(b)), jnp.sign(b))

    assert jnp.isclose((_a + _b).value, a + b)
    assert jnp.isclose((_a - _b).value, a - b)
    assert jnp.isclose((_a * _b).value, a * b)
    assert jnp.isclose((_a / _b).value, a / b)
    assert jnp.isclose((_a ** 2).value, a ** 2)
    assert jnp.isclose((_a ** 3).value, a ** 3)

    a = 0.
    b = 10.

    _a = LogSpace(jnp.log(jnp.abs(a)), jnp.sign(a))
    _b = LogSpace(jnp.log(jnp.abs(b)), jnp.sign(b))

    assert jnp.isclose((_a + _b).value, a + b)
    assert jnp.isclose((_a - _b).value, a - b)
    assert jnp.isclose((_a * _b).value, a * b)
    assert jnp.isclose((_a / _b).value, a / b)
    assert jnp.isclose((_a ** 2).value, a ** 2)
    assert jnp.isclose((_a ** 3).value, a ** 3)

    a = jnp.arange(5)
    sign = jnp.sign(a)
    _a = LogSpace(jnp.log(a), sign)
    assert jnp.allclose(_a[:2].value, a[:2])


def test_signed_logaddexp():
    for i in range(100):
        u = random.uniform(random.PRNGKey(i), shape=(2,)) * 20. - 10.
        a = jnp.log(jnp.abs(u[0]))
        b = jnp.log(jnp.abs(u[1]))
        sign1 = jnp.sign(u[0])
        sign2 = jnp.sign(u[1])
        ans = u[0] + u[1]
        ans_sign = jnp.sign(ans)
        log_abs_ans = jnp.log(jnp.abs(ans))
        log_abs_c, sign_c = signed_logaddexp(a, sign1, b, sign2)
        assert sign_c == ans_sign
        assert jnp.isclose(log_abs_c, log_abs_ans)

    u = [1., 1.]
    a = jnp.log(jnp.abs(u[0]))
    b = jnp.log(jnp.abs(u[1]))
    sign1 = jnp.sign(u[0])
    sign2 = jnp.sign(u[1])
    ans = u[0] + u[1]
    ans_sign = jnp.sign(ans)
    log_abs_ans = jnp.log(jnp.abs(ans))
    log_abs_c, sign_c = signed_logaddexp(a, sign1, b, sign2)
    assert sign_c == ans_sign
    assert jnp.isclose(log_abs_c, log_abs_ans)

    u = [1., -1.]
    a = jnp.log(jnp.abs(u[0]))
    b = jnp.log(jnp.abs(u[1]))
    sign1 = jnp.sign(u[0])
    sign2 = jnp.sign(u[1])
    ans = u[0] + u[1]
    ans_sign = jnp.sign(ans)
    log_abs_ans = jnp.log(jnp.abs(ans))
    log_abs_c, sign_c = signed_logaddexp(a, sign1, b, sign2)
    # assert sign_c == ans_sign
    assert jnp.isclose(log_abs_c, log_abs_ans)

    u = [-1., 1.]
    a = jnp.log(jnp.abs(u[0]))
    b = jnp.log(jnp.abs(u[1]))
    sign1 = jnp.sign(u[0])
    sign2 = jnp.sign(u[1])
    ans = u[0] + u[1]
    ans_sign = jnp.sign(ans)
    log_abs_ans = jnp.log(jnp.abs(ans))
    log_abs_c, sign_c = signed_logaddexp(a, sign1, b, sign2)
    # assert sign_c == ans_sign
    assert jnp.isclose(log_abs_c, log_abs_ans)

    u = [0., 0.]
    a = jnp.log(jnp.abs(u[0]))
    b = jnp.log(jnp.abs(u[1]))
    sign1 = jnp.sign(u[0])
    sign2 = jnp.sign(u[1])
    ans = u[0] + u[1]
    ans_sign = jnp.sign(ans)
    log_abs_ans = jnp.log(jnp.abs(ans))
    log_abs_c, sign_c = signed_logaddexp(a, sign1, b, sign2)
    assert sign_c == ans_sign
    assert jnp.isclose(log_abs_c, log_abs_ans)

    u = [0., 1.]
    a = jnp.log(jnp.abs(u[0]))
    b = jnp.log(jnp.abs(u[1]))
    sign1 = jnp.sign(u[0])
    sign2 = jnp.sign(u[1])
    ans = u[0] + u[1]
    ans_sign = jnp.sign(ans)
    log_abs_ans = jnp.log(jnp.abs(ans))
    log_abs_c, sign_c = signed_logaddexp(a, sign1, b, sign2)
    assert sign_c == ans_sign
    assert jnp.isclose(log_abs_c, log_abs_ans)

    u = [0., -1.]
    a = jnp.log(jnp.abs(u[0]))
    b = jnp.log(jnp.abs(u[1]))
    sign1 = jnp.sign(u[0])
    sign2 = jnp.sign(u[1])
    ans = u[0] + u[1]
    ans_sign = jnp.sign(ans)
    log_abs_ans = jnp.log(jnp.abs(ans))
    log_abs_c, sign_c = signed_logaddexp(a, sign1, b, sign2)
    assert sign_c == ans_sign
    assert jnp.isclose(log_abs_c, log_abs_ans)


def test_cumulative_logsumexp():
    a = jnp.linspace(-1., 1., 100)
    v1 = jnp.log(jnp.cumsum(jnp.exp(a)))
    v2 = cumulative_logsumexp(a)
    assert jnp.allclose(v1, v2)


def test_logaddexp():
    a = jnp.log(1.)
    b = jnp.log(1.)
    assert logaddexp(a, b) == jnp.log(2.)
    a = jnp.log(1.)
    b = jnp.log(-2. + 0j)
    assert jnp.isclose(jnp.exp(logaddexp(a, b)).real, -1.)

    a = jnp.log(-1. + 0j)
    b = jnp.log(2. + 0j)
    assert jnp.isclose(jnp.exp(logaddexp(a, b)).real, 1.)

    for i in range(100):
        u = random.uniform(random.PRNGKey(i), shape=(2,)) * 20. - 10.
        a = jnp.log(u[0] + 0j)
        b = jnp.log(u[1] + 0j)
        assert jnp.isclose(jnp.exp(logaddexp(a, b)).real, u[0] + u[1], atol=5e-5)


def test_is_complex():
    assert is_complex(jnp.ones(1, dtype=complex_type))



from abc import abstractmethod
from typing import Tuple, Optional, Generator, Callable

import jax.numpy as jnp

from jaxns.framework.abc import AbstractModel, AbstractPrior, AbstractDistribution
from jaxns.internals.shapes import tuple_prod
from jaxns.internals.types import LikelihoodInputType
from jaxns.internals.types import LikelihoodType, UType, XType, RandomVariableType, MeasureType

__all__ = [
    'BaseAbstractModel',
    "BaseAbstractPrior",
    "BaseAbstractDistribution",
    "PriorModelGen",
    "PriorModelType"
]


class BaseAbstractPrior(AbstractPrior):
    """
    The base prior class with public methods.
    """

    def __init__(self, name: Optional[str] = None):
        self.name = name

    def __repr__(self):
        return f"{self.name if self.name is not None else '*'}\t{self.base_shape} -> {self.shape} {self.dtype}"

    @property
    def dtype(self):
        """
        The dtype of the prior random variable in X-space.
        """
        return self._dtype()

    @property
    def base_shape(self) -> Tuple[int, ...]:
        """
        The base shape of the prior random variable in U-space.
        """
        return self._base_shape()

    @property
    def base_ndims(self):
        """
        The number of dimensions of the prior random variable in U-space.
        """
        return tuple_prod(self.base_shape)

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        The shape of the prior random variable in X-space.
        """
        return self._shape()

    def forward(self, U: UType) -> RandomVariableType:
        """
        The forward transformation from U-space to X-space.

        Args:
            U: U-space representation

        Returns:
            X-space representation
        """
        return self._forward(U)

    def inverse(self, X: RandomVariableType) -> UType:
        """
        The inverse transformation from X-space to U-space.

        Args:
            X: X-space representation

        Returns:
            U-space representation
        """
        return self._inverse(X)

    def log_prob(self, X: RandomVariableType) -> MeasureType:
        """
        The log probability of the prior.

        Args:
            X: X-space representation

        Returns:
            log probability of the prior
        """
        log_prob = self._log_prob(X)
        if log_prob.size > 1:
            log_prob = jnp.sum(log_prob)
        if log_prob.shape != ():
            log_prob = log_prob.reshape(())
        return log_prob


PriorModelGen = Generator[BaseAbstractPrior, RandomVariableType, LikelihoodInputType]
PriorModelType = Callable[[], PriorModelGen]


class BaseAbstractModel(AbstractModel):
    """
    The base model class with public methods.
    """

    def __init__(self, prior_model: PriorModelType, log_likelihood: LikelihoodType):
        self._prior_model = prior_model
        self._log_likelihood = log_likelihood

    @property
    def prior_model(self) -> PriorModelType:
        """
        The prior model.
        """
        return self._prior_model

    @property
    def log_likelihood(self) -> LikelihoodType:
        """
        The log likelihood function.

        Returns:
            log likelihood function
        """
        return self._log_likelihood

    @abstractmethod
    def _U_placeholder(self) -> UType:
        ...

    @abstractmethod
    def _X_placeholder(self) -> XType:
        ...

    @property
    def U_placeholder(self) -> UType:
        """
        A placeholder for U-space sample.
        """
        return self._U_placeholder()

    @property
    def X_placeholder(self) -> XType:
        """
        A placeholder for X-space sample.
        """
        return self._X_placeholder()

    @property
    def U_ndims(self) -> int:
        """
        The prior dimensionality.
        """
        return self.U_placeholder.size


# TODO(Joshuaalbert): distribution is too similar to prior, where we only need to be able to extract the log_prob and
#  potential tranformations from the underlying. Perhaps we should just define as ops to create priors. Try making
#  priors just use distribution functionality. Special priors will need treatment.
class BaseAbstractDistribution(AbstractDistribution):
    """
    The base distribution class with public methods.
    """

    @property
    def dtype(self):
        """
        The dtype of the distribution, in X-space.
        """
        return self._dtype()

    @property
    def base_shape(self) -> Tuple[int, ...]:
        """
        The base shape of the distribution, in U-space.
        """
        return self._base_shape()

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        The shape of the distribution, in X-space.
        """
        return self._shape()

    def forward(self, U: UType) -> RandomVariableType:
        """
        The forward transformation from U-space to X-space.

        Args:
            U: U-space representation

        Returns:
            X-space representation
        """
        return self._forward(U)

    def inverse(self, X: RandomVariableType) -> UType:
        """
        The inverse transformation from X-space to U-space.

        Args:
            X: X-space representation

        Returns:
            U-space representation
        """
        return self._inverse(X)

    def log_prob(self, X: RandomVariableType) -> MeasureType:
        """
        The log probability of the distribution.

        Args:
            X: X-space representation

        Returns:
            log probability of the distribution
        """
        return self._log_prob(X)



import logging
from typing import TypeVar, NamedTuple, Tuple

import jax
from jax import numpy as jnp, random, lax, tree_map

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.cumulative_ops import cumulative_op_static
from jaxns.internals.types import PRNGKey, FloatArray, BoolArray, Sample, float_type, int_type, \
    StaticStandardNestedSamplerState, \
    IntArray, UType, StaticStandardSampleCollection
from jaxns.samplers.abc import SamplerState
from jaxns.samplers.bases import SeedPoint, BaseAbstractMarkovSampler

__all__ = [
    'UniDimSliceSampler'
]

logger = logging.getLogger('jaxns')

T = TypeVar('T')


def _sample_direction(n_key: PRNGKey, ndim: int) -> FloatArray:
    """
    Choose a direction randomly from S^(D-1).

    Args:
        n_key: PRNG key
        ndim: int, number of dimentions

    Returns:
        direction: [D] direction from S^(D-1)
    """
    if ndim == 1:
        return jnp.ones(())
    direction = random.normal(n_key, shape=(ndim,), dtype=float_type)
    direction /= jnp.linalg.norm(direction)
    return direction


def _slice_bounds(point_U0: FloatArray, direction: FloatArray) -> Tuple[FloatArray, FloatArray]:
    """
    Compute the slice bounds, t, where point_U0 + direction * t intersects uit cube boundary.

    Args:
        point_U0: [D]
        direction: [D]

    Returns:
        left_bound: left most point (<= 0).
        right_bound: right most point (>= 0).
    """
    t1 = (1. - point_U0) / direction
    t1_right = jnp.min(jnp.where(t1 >= 0., t1, jnp.inf))
    t1_left = jnp.max(jnp.where(t1 <= 0., t1, -jnp.inf))
    t0 = -point_U0 / direction
    t0_right = jnp.min(jnp.where(t0 >= 0., t0, jnp.inf))
    t0_left = jnp.max(jnp.where(t0 <= 0., t0, -jnp.inf))
    right_bound = jnp.minimum(t0_right, t1_right)
    left_bound = jnp.maximum(t0_left, t1_left)
    return left_bound, right_bound


def _pick_point_in_interval(key: PRNGKey, point_U0: FloatArray, direction: FloatArray, left: FloatArray,
                            right: FloatArray) -> Tuple[FloatArray, FloatArray]:
    """
    Select a point along slice in [point_U0 + direction * left, point_U0 + direction * right]

    Args:
        key: PRNG key
        point_U0: [D]
        direction: [D]
        left: left most point (<= 0).
        right: right most point (>= 0).

    Returns:
        point_U: [D]
        t: selection point between [left, right]
    """
    t = random.uniform(key, minval=left, maxval=right, dtype=float_type)
    point_U = point_U0 + t * direction
    # close_to_zero = (left >= -10*jnp.finfo(left.dtype).eps) & (right <= 10*jnp.finfo(right.dtype).eps)
    # point_U = jnp.where(close_to_zero, point_U0, point_U)
    # t = jnp.where(close_to_zero, jnp.zeros_like(t), t)
    return point_U, t


def _shrink_interval(key: PRNGKey, t: FloatArray, left: FloatArray, right: FloatArray, log_L_proposal: FloatArray,
                     log_L_constraint: FloatArray, log_L0: FloatArray,
                     midpoint_shrink: bool) -> Tuple[FloatArray, FloatArray]:
    """
    Not successful proposal, so shrink, optionally apply exponential shrinkage.
    """
    # witout exponential shrinkage, we shrink to failed proposal point, which is 100% correct.
    left = jnp.where(t < 0., t, left)
    right = jnp.where(t > 0., t, right)
    key, t_key, midpoint_key = random.split(key, 3)

    if midpoint_shrink:
        # we take two points along lines from origin:
        #  - alpha: one from a satisfying point (t=0) to non-satisfying proposal,
        #  - beta: and, one from non-satisfying proposal to the constraint.
        # We shrink to alpha point if beta point is above alpha point.
        # Intuitively, the tangent from constraint is more accurate than proposal and should be a supremum of heights
        # in reasonable cases.
        # An extension is to make alpha shrink to constraint line, which would shrink very fast, but introduce
        # auto-correlation which must be later refined away.
        # Line logic:
        # logL(t) = m * t + b
        # logL(0) = b
        # (logL(t_R) - logL(0))/t_R
        # logL(t_R*alpha) = (logL(t_R) - logL(0))*alpha + logL(0)
        alpha_key, beta_key = random.split(midpoint_key, 2)
        alpha = random.uniform(alpha_key)
        beta = random.uniform(beta_key)
        logL_alpha = log_L0 + alpha * (log_L_proposal - log_L0)
        logL_beta = log_L_proposal + beta * (log_L_constraint - log_L_proposal)
        do_mid_point_shrink = logL_alpha < logL_beta
        left = jnp.where((t < 0.) & do_mid_point_shrink, alpha * left, left)
        right = jnp.where((t > 0.) & do_mid_point_shrink, alpha * right, right)
    return left, right


def _new_proposal(key: PRNGKey, seed_point: SeedPoint, midpoint_shrink: bool, perfect: bool,
                  gradient_slice: bool,
                  log_L_constraint: FloatArray,
                  model: BaseAbstractModel) -> Tuple[FloatArray, FloatArray, IntArray]:
    """
    Sample from a slice about a seed point.

    Args:
        key: PRNG key
        seed_point: the seed point to sample from
        midpoint_shrink: if true then contract to the midpoint of interval on rejection. Otherwise, normal contract
        perfect: if true then perform exponential shrinkage from maximal bounds, requiring no step-out procedure.
        gradient_slice: if true the slice along gradient direction
        log_L_constraint: the constraint to sample within
        model: the model to sample from

    Returns:
        point_U: the new sample
        log_L: the log-likelihood of the new sample
        num_likelihood_evaluations: the number of likelihood evaluations performed
    """

    class Carry(NamedTuple):
        key: PRNGKey
        direction: FloatArray
        left: FloatArray
        right: FloatArray
        t: FloatArray
        point_U: UType
        log_L: FloatArray
        num_likelihood_evaluations: IntArray

    def cond(carry: Carry) -> BoolArray:
        close_to_zero_interval = (carry.right - carry.left) <= 2 * jnp.finfo(carry.right.dtype).eps
        satisfaction = carry.log_L > log_L_constraint
        # Allow if on plateau to fly around the plateau for a while
        lesser_satisfaction = jnp.bitwise_and(seed_point.log_L0 == log_L_constraint, carry.log_L == log_L_constraint)
        done = jnp.bitwise_or(jnp.bitwise_or(close_to_zero_interval, satisfaction), lesser_satisfaction)
        return jnp.bitwise_not(done)

    def body(carry: Carry) -> Carry:
        key, t_key, shrink_key = random.split(carry.key, 3)
        left, right = _shrink_interval(
            key=shrink_key,
            t=carry.t,
            left=carry.left,
            right=carry.right,
            log_L_proposal=carry.log_L,
            log_L_constraint=log_L_constraint,
            log_L0=seed_point.log_L0,
            midpoint_shrink=midpoint_shrink
        )
        point_U, t = _pick_point_in_interval(
            key=t_key,
            point_U0=seed_point.U0,
            direction=carry.direction,
            left=left,
            right=right
        )
        log_L = model.forward(point_U)
        num_likelihood_evaluations = carry.num_likelihood_evaluations + jnp.ones_like(carry.num_likelihood_evaluations)
        return Carry(
            key=key,
            t=t,
            left=left,
            right=right,
            point_U=point_U,
            log_L=log_L,
            num_likelihood_evaluations=num_likelihood_evaluations,
            direction=carry.direction
        )

    key, n_key, t_key = random.split(key, 3)
    if gradient_slice:
        direction = jax.grad(model.forward)(seed_point.U0)
        norm = jnp.linalg.norm(direction)
        direction /= norm
        direction = jnp.where(
            jnp.bitwise_or(jnp.equal(norm, jnp.zeros_like(norm)), ~jnp.isfinite(norm)),
            _sample_direction(n_key, seed_point.U0.size),
            direction
        )
        num_likelihood_evaluations = jnp.full((), 1, int_type)
        (left, right) = _slice_bounds(
            point_U0=seed_point.U0,
            direction=direction
        )
        left = jnp.zeros_like(left)
    else:
        direction = _sample_direction(n_key, seed_point.U0.size)
        num_likelihood_evaluations = jnp.full((), 0, int_type)
        if perfect:
            (left, right) = _slice_bounds(
                point_U0=seed_point.U0,
                direction=direction
            )
        else:
            # TODO: implement doubling step out
            raise NotImplementedError("TODO: implement doubling step out")
    point_U, t = _pick_point_in_interval(
        key=t_key,
        point_U0=seed_point.U0,
        direction=direction,
        left=left,
        right=right
    )
    log_L = model.forward(point_U)
    init_carry = Carry(
        key=key,
        direction=direction,
        left=left,
        right=right,
        t=t,
        point_U=point_U,
        log_L=log_L,
        num_likelihood_evaluations=num_likelihood_evaluations
    )

    carry = lax.while_loop(
        cond_fun=cond,
        body_fun=body,
        init_val=init_carry
    )
    return carry.point_U, carry.log_L, carry.num_likelihood_evaluations


class UniDimSliceSampler(BaseAbstractMarkovSampler):
    """
    Slice sampler for a single dimension. Produces correlated (non-i.i.d.) samples.
    """

    def __init__(self, model: BaseAbstractModel, num_slices: int, num_phantom_save: int, midpoint_shrink: bool,
                 perfect: bool, gradient_slice: bool = False):
        """
        Unidimensional slice sampler.

        Args:
            model: AbstractModel
            num_slices: number of slices between acceptance. Note: some other software use units of prior dimension.
            midpoint_shrink: if true then contract to the midpoint of interval on rejection. Otherwise, contract to
                rejection point. Speeds up convergence, but introduces minor auto-correlation.
            num_phantom_save: number of phantom samples to save. Phantom samples are samples that meeting the constraint
                but are not accepted. They can be used for numerous things, e.g. to estimate the evidence uncertainty.
            perfect: if true then perform exponential shrinkage from maximal bounds, requiring no step-out procedure.
                Otherwise, uses a doubling procedure (exponentially finding bracket).
                Note: Perfect is a misnomer, as perfection also depends on the number of slices between acceptance.
            gradient_slice: if true then always slice along gradient direction.
        """
        super().__init__(model=model)
        if num_slices < 1:
            raise ValueError(f"num_slices should be >= 1, got {num_slices}.")
        if num_phantom_save < 0:
            raise ValueError(f"num_phantom_save should be >= 0, got {num_phantom_save}.")
        if num_phantom_save >= num_slices:
            raise ValueError(f"num_phantom_save should be < num_slices, got {num_phantom_save} >= {num_slices}.")
        self.num_slices = int(num_slices)
        self.num_phantom_save = int(num_phantom_save)
        self.midpoint_shrink = bool(midpoint_shrink)
        self.perfect = bool(perfect)
        self.gradient_slice = bool(gradient_slice)
        if not self.perfect:
            raise ValueError("Only perfect slice sampler is implemented.")

    def num_phantom(self) -> int:
        return self.num_phantom_save

    def pre_process(self, state: StaticStandardNestedSamplerState) -> SamplerState:
        sample_collection = tree_map(lambda x: x[state.front_idx], state.sample_collection)
        if self.perfect:  # nothing needed
            return (sample_collection,)
        else:  # TODO: step out with doubling, using ellipsoidal clustering
            return (sample_collection,)  # multi_ellipsoidal_params()

    def post_process(self, sample_collection: StaticStandardSampleCollection,
                     sampler_state: SamplerState) -> SamplerState:
        if self.perfect:  # nothing needed
            return (sample_collection,)
        else:  # TODO: step out with doubling, using ellipsoidal clustering, could shrink ellipsoids
            return (sample_collection,)

    def get_seed_point(self, key: PRNGKey, sampler_state: SamplerState,
                       log_L_constraint: FloatArray) -> SeedPoint:

        sample_collection: StaticStandardSampleCollection
        (sample_collection,) = sampler_state

        select_mask = sample_collection.log_L > log_L_constraint
        # If non satisfied samples, then choose randomly from them.
        any_satisfied = jnp.any(select_mask)
        yes_ = jnp.asarray(0., float_type)
        no_ = jnp.asarray(-jnp.inf, float_type)
        unnorm_select_log_prob = jnp.where(
            any_satisfied,
            jnp.where(select_mask, yes_, no_),
            yes_
        )
        # Choose randomly where mask is True
        g = random.gumbel(key, shape=unnorm_select_log_prob.shape)
        sample_idx = jnp.argmax(g + unnorm_select_log_prob)

        return SeedPoint(
            U0=sample_collection.U_samples[sample_idx],
            log_L0=sample_collection.log_L[sample_idx]
        )

    def get_sample_from_seed(self, key: PRNGKey, seed_point: SeedPoint, log_L_constraint: FloatArray,
                             sampler_state: SamplerState) -> Tuple[Sample, Sample]:

        def propose_op(sample: Sample, key: PRNGKey) -> Sample:
            U_sample, log_L, num_likelihood_evaluations = _new_proposal(
                key=key,
                seed_point=SeedPoint(
                    U0=sample.U_sample,
                    log_L0=sample.log_L
                ),
                midpoint_shrink=self.midpoint_shrink,
                perfect=self.perfect,
                gradient_slice=self.gradient_slice,
                log_L_constraint=log_L_constraint,
                model=self.model
            )
            return Sample(
                U_sample=U_sample,
                log_L_constraint=log_L_constraint,
                log_L=log_L,
                num_likelihood_evaluations=num_likelihood_evaluations + sample.num_likelihood_evaluations
            )

        init_sample = Sample(
            U_sample=seed_point.U0,
            log_L_constraint=log_L_constraint,
            log_L=seed_point.log_L0,
            num_likelihood_evaluations=jnp.asarray(0, int_type)
        )
        final_sample, cumulative_samples = cumulative_op_static(
            op=propose_op,
            init=init_sample,
            xs=random.split(key, self.num_slices),
            unroll=2
        )

        # Last sample is the final sample, the rest are potential phantom samples
        # Take only the last num_phantom_save phantom samples
        phantom_samples: Sample = tree_map(lambda x: x[-(self.num_phantom_save + 1):-1], cumulative_samples)

        # Due to the cumulative nature of the sampler, the final number of likelihood evaluations should be divided
        # equally among the accepted sample and retained phantom samples.
        num_likelihood_evaluations_per_phantom_sample = (
                final_sample.num_likelihood_evaluations / (self.num_phantom_save + 1)
        ).astype(int_type)
        num_likelihood_evaluations_per_accepted_sample = (
                final_sample.num_likelihood_evaluations - num_likelihood_evaluations_per_phantom_sample * self.num_phantom_save
        )
        final_sample = final_sample._replace(
            num_likelihood_evaluations=num_likelihood_evaluations_per_accepted_sample
        )
        phantom_samples = phantom_samples._replace(
            num_likelihood_evaluations=jnp.full(
                phantom_samples.num_likelihood_evaluations.shape,
                num_likelihood_evaluations_per_phantom_sample,
                phantom_samples.num_likelihood_evaluations.dtype
            )
        )
        return final_sample, phantom_samples



from abc import abstractmethod
from typing import NamedTuple, Tuple

from jax import random

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.types import FloatArray, Sample
from jaxns.internals.types import PRNGKey
from jaxns.samplers.abc import AbstractSampler, SamplerState


class BaseAbstractSampler(AbstractSampler):
    def __init__(self, model: BaseAbstractModel):
        self.model = model

    def __repr__(self):
        return f"{self.__class__.__name__}"


class BaseAbstractRejectionSampler(BaseAbstractSampler):
    """
    Samplers that are based on rejection sampling. They usually first-lines of attack, and are stopped once efficiency
    gets too low.
    """
    pass


class SeedPoint(NamedTuple):
    U0: FloatArray
    log_L0: FloatArray


class BaseAbstractMarkovSampler(BaseAbstractSampler):
    """
    A sampler that conditions off a known satisfying point, e.g. a seed point.
    """

    @abstractmethod
    def get_sample_from_seed(self, key: PRNGKey, seed_point: SeedPoint, log_L_constraint: FloatArray,
                             sampler_state: SamplerState) -> Tuple[Sample, Sample]:
        """
        Produce a single i.i.d. sample from the model within the log_L_constraint.

        Args:
            key: PRNGkey
            seed_point: function that gets the next sample from a seed point
            log_L_constraint: the constraint to sample within
            sampler_state: the data pytree needed and produced by the sampler

        Returns:
            an i.i.d. sample, and batched phantom samples
        """
        ...

    @abstractmethod
    def get_seed_point(self, key: PRNGKey, sampler_state: SamplerState,
                       log_L_constraint: FloatArray) -> SeedPoint:
        """
        Samples a seed point from the live points.

        Args:
            key: PRNGKey
            sampler_state: the current sampler state
            log_L_constraint: a log-L constraint to sample within. Must always be at least one sample in front above
                this to avoid infinite loop.

        Returns:
            a seed point
        """
        ...

    def get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, sampler_state: SamplerState) -> Tuple[
        Sample, Sample]:
        sample_key, seed_key = random.split(key, 2)
        seed_point = self.get_seed_point(
            key=seed_key,
            sampler_state=sampler_state,
            log_L_constraint=log_L_constraint
        )
        return self.get_sample_from_seed(
            key=sample_key,
            seed_point=seed_point,
            log_L_constraint=log_L_constraint,
            sampler_state=sampler_state
        )



import jax
from jax import numpy as jnp

from jaxns.internals.maps import replace_index, chunked_pmap, prepare_func_args, get_index, chunked_vmap


def test_replace_index():
    operand = jnp.asarray([0, 1, 2, 3, 4])
    update = jnp.asarray([5, 5])
    start_idx = 0
    expect = jnp.asarray([5, 5, 2, 3, 4])
    assert jnp.all(replace_index(operand, update, start_idx) == expect)

    operand = jnp.asarray([0, 1, 2, 3, 4])
    update = jnp.asarray(5)
    start_idx = 0
    expect = jnp.asarray([5, 1, 2, 3, 4])
    assert jnp.all(replace_index(operand, update, start_idx) == expect)

    operand = jnp.asarray([0, 1, 2, 3, 4])
    update = jnp.asarray([5, 5])
    start_idx = 4
    expect = jnp.asarray([0, 1, 2, 5, 5])
    assert jnp.all(replace_index(operand, update, start_idx) == expect)

    operand = jnp.asarray([0, 1, 2, 3, 4])
    update = jnp.asarray(5)
    start_idx = 4
    expect = jnp.asarray([0, 1, 2, 3, 5])
    assert jnp.all(replace_index(operand, update, start_idx) == expect)


def test_chunked_pmap():
    def f(x, y):
        return x * y

    chunked_f = chunked_pmap(f, 1)
    x = jnp.arange(3)
    assert chunked_f(x, y=x).shape == x.shape
    assert jnp.all(chunked_f(x, y=x) == x ** 2)

    num_dev = len(jax.devices())
    chunked_f = chunked_pmap(f, num_dev)
    x = jnp.arange(num_dev)
    assert chunked_f(x, y=x).shape == x.shape
    assert jnp.all(chunked_f(x, y=x) == x ** 2)

    x = jnp.arange(3)
    assert chunked_f(x, y=x).shape == x.shape
    assert jnp.all(chunked_f(x, y=x) == x ** 2)


def test_prepare_func_args():
    prepare_func_args(lambda a: a)(a=1, b=2)

    def f(a, b=1):
        return a + b

    g = prepare_func_args(f)
    kwargs = dict(a=1, b=2, c=3)
    assert g(**kwargs) == f(kwargs['a'], b=kwargs['b'])
    kwargs = dict(a=1, c=3)
    assert g(**kwargs) == f(kwargs['a'])

    def f(a, b=2, *, c, d=4):
        return a + b + c + d

    g = prepare_func_args(f)
    kwargs = dict(a=5, b=6, c=7, d=8)
    assert g(**kwargs) == f(kwargs['a'], b=kwargs['b'], c=kwargs['c'], d=kwargs['d'])
    kwargs = dict(a=9, c=11)
    assert g(**kwargs) == f(kwargs['a'], c=kwargs['c'])


def test_get_index():
    operand = jnp.asarray([[1, 2, 3], [4, 5, 6]])
    start_index = 0
    length = 1
    expect = jnp.asarray([[1, 2, 3]])
    assert jnp.allclose(get_index(operand, start_index, length), expect)

    operand = jnp.asarray([[1, 2, 3], [4, 5, 6]])
    start_index = 0
    length = 2
    expect = operand
    assert jnp.allclose(get_index(operand, start_index, length), expect)

    operand = jnp.asarray([[1, 2, 3], [4, 5, 6]])
    start_index = 1
    length = 2
    expect = operand
    assert jnp.allclose(get_index(operand, start_index, length), expect)

    operand = jnp.asarray([[1, 2, 3], [4, 5, 6]])
    start_index = 1
    length = 1
    expect = jnp.asarray([[4, 5, 6]])
    assert jnp.allclose(get_index(operand, start_index, length), expect)


def test_chunked_vmap():
    def f(x, y):
        return x * y

    chunked_f = chunked_vmap(f, chunk_size=2)
    x = jnp.arange(3)
    assert chunked_f(x, y=x).shape == x.shape
    assert jnp.all(chunked_f(x, y=x) == x ** 2)

    chunked_f = chunked_vmap(f, chunk_size=1)
    x = jnp.arange(3)
    assert chunked_f(x, y=x).shape == x.shape
    assert jnp.all(chunked_f(x, y=x) == x ** 2)

    chunked_f = chunked_vmap(f, chunk_size=2)
    x = jnp.arange(6)
    assert chunked_f(x, y=x).shape == x.shape
    assert jnp.all(chunked_f(x, y=x) == x ** 2)

    chunked_f = chunked_vmap(f, chunk_size=1)
    x = jnp.arange(6)
    assert chunked_f(x, y=x).shape == x.shape
    assert jnp.all(chunked_f(x, y=x) == x ** 2)
